"""15 hatch, 14 gas, 14 pool ling flood."""
from typing import List, Union

from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId
from sc2.player import Race

from sharpy.general.zone import Zone
from sharpy.managers.roles import UnitTask
from sharpy.plans import BuildOrder, SequentialList, Step, StepBuildGas
from sharpy.plans.acts import Tech, ActBuilding, ActBase, DefensePosition, DefensiveBuilding
from sharpy.plans.acts.zerg import ZergUnit, AutoOverLord, MorphLair
from sharpy.plans.acts import Expand
from sharpy.plans.require import (
    UnitExists,
    Supply,
    All,
    Gas,
    Minerals,
    TechReady,
    Any,
)
from sharpy.plans.tactics import (
    PlanDistributeWorkers,
    PlanZoneAttackAllIn,
    PlanWorkerOnlyDefense,
    PlanZoneDefense,
    PlanFinishEnemy,
)
from paul_plans.operations import EXTRACTOR_TRICK, CustomInject


class CustomCounterTerranTie(BuildOrder):
    def __init__(self, orders: List[Union[ActBase, List[ActBase]]]):
        """
        Build order package that replaces normal build order for Zerg with one that builds mutalisks to destroy terran
        flying buildings.
        Add any PlanDistributeWorkers acts with orders
        """
        cover_list = SequentialList(
            [
                PlanDistributeWorkers(),
                AutoOverLord(),
                Step(None, ZergUnit(UnitTypeId.DRONE, 20), skip=Supply(198)),
                StepBuildGas(4, None),
                MorphLair(),
                ActBuilding(UnitTypeId.SPIRE, 1),
                Step(
                    None,
                    DefensiveBuilding(UnitTypeId.SPORECRAWLER, DefensePosition.BehindMineralLineCenter),
                    skip_until=Supply(199),
                ),
                Step(
                    None, DefensiveBuilding(UnitTypeId.SPINECRAWLER, DefensePosition.Entrance), skip_until=Supply(199),
                ),
                ZergUnit(UnitTypeId.MUTALISK, 10),
            ]
        )

        new_build_order = [
            Step(None, cover_list, skip_until=self.should_build_mutalisks),
            Step(None, BuildOrder(orders), skip=self.should_build_mutalisks),
        ]
        super().__init__(new_build_order)

    def should_build_mutalisks(self, _knowledge):
        #  knowledge isn't used but it gets passed to the function because of skip/skip_until coding
        if self.knowledge.enemy_race != Race.Terran:
            return False

        if len(self.cache.own({UnitTypeId.MUTALISK, UnitTypeId.CORRUPTOR})) >= 10:
            return False

        if len(self.ai.enemy_units.not_flying) > 1:
            return False

        if self.ai.supply_used < 70:
            return False

        main_zone: Zone = self.knowledge.enemy_main_zone
        if not main_zone.is_scouted_at_least_once:
            return False

        buildings = self.ai.enemy_structures
        return len(buildings) == len(buildings.flying)


class LingRush(BuildOrder):
    """
    Original Ling Rush for Paul.

    15 hatch/14 gas/14 pool > ling flood. No transition intended.
    """

    def __init__(self):
        """Build order steps to be combined."""
        self.rally_set = False
        self.attack = PlanZoneAttackAllIn()

        attack_tactics = [
            PlanWorkerOnlyDefense(),
            self.attack,  # PlanZoneAttackAllIn
            PlanZoneDefense(),
            PlanFinishEnemy(),
        ]

        opening = SequentialList(
            [
                Step(None, ZergUnit(UnitTypeId.DRONE, to_count=14)),
                Step(
                    Supply(14),
                    EXTRACTOR_TRICK,
                    skip=UnitExists(UnitTypeId.EXTRACTOR, 1, include_killed=True, include_not_ready=False,),
                ),
                Expand(2, priority=True, consider_worker_production=False),
                Step(All([Supply(14), Minerals(130)]), StepBuildGas(1),),
                Step(None, ZergUnit(UnitTypeId.DRONE, to_count=14)),
                Step(Supply(14), ActBuilding(UnitTypeId.SPAWNINGPOOL)),
                Step(
                    All([UnitExists(UnitTypeId.HATCHERY), UnitExists(UnitTypeId.SPAWNINGPOOL)]),
                    ZergUnit(UnitTypeId.QUEEN, 2),
                ),
                # spam lings
                Step(UnitExists(UnitTypeId.SPAWNINGPOOL), ZergUnit(UnitTypeId.ZERGLING, 999)),
            ]
        )

        ling_speed = BuildOrder(
            [
                Step(
                    All([UnitExists(UnitTypeId.SPAWNINGPOOL), Gas(100)]),
                    Tech(UpgradeId.ZERGLINGMOVEMENTSPEED, UnitTypeId.SPAWNINGPOOL),
                )
            ]
        )

        worker_distribution = SequentialList(
            [
                Step(
                    Supply(15), PlanDistributeWorkers(), skip=UnitExists(UnitTypeId.HATCHERY, 2, include_killed=True,),
                ),
                Step(
                    UnitExists(UnitTypeId.EXTRACTOR),
                    PlanDistributeWorkers(min_gas=3),
                    skip=Any([TechReady(UpgradeId.ZERGLINGMOVEMENTSPEED, percentage=0.01), Gas(96)]),
                ),
                Step(
                    Any([TechReady(UpgradeId.ZERGLINGMOVEMENTSPEED, percentage=0.01), Gas(96)]),
                    PlanDistributeWorkers(max_gas=0),
                ),
            ]
        )

        overlords = BuildOrder([Step(None, AutoOverLord(), skip_until=UnitExists(UnitTypeId.SPAWNINGPOOL),)])

        super().__init__(
            CustomCounterTerranTie([CustomInject(), opening, ling_speed, worker_distribution, overlords]),
            attack_tactics,
        )

    async def execute(self) -> bool:
        if not self.rally_set and self.get_count(UnitTypeId.HATCHERY) == 2:
            for hatch in self.cache.own(UnitTypeId.HATCHERY):
                self.ai.do(hatch(AbilityId.RALLY_HATCHERY_UNITS, self.knowledge.expansion_zones[1].gather_point))
            self.rally_set = True
        if UpgradeId.ZERGLINGMOVEMENTSPEED in self.ai.state.upgrades:
            # attacking_lings = self.roles.get_types_from({UnitTypeId.ZERGLING}, UnitTask.Attacking)
            # moving_lings = self.roles.get_types_from({UnitTypeId.ZERGLING}, UnitTask.Moving)
            idle_lings = self.roles.get_types_from({UnitTypeId.ZERGLING}, UnitTask.Idle)
            for unit in self.cache.own(UnitTypeId.ZERGLING):
                if unit not in idle_lings:  # attacking_lings | moving_lings:
                    self.combat.add_unit(unit)
            self.attack.start_attack_power = 1
            self.attack.attack_on_advantage = False
            self.attack.retreat_multiplier = 0
        return await super(LingRush, self).execute()
