from typing import TYPE_CHECKING
from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId
from sharpy.managers.roles import UnitTask
from sharpy.managers.combat2 import MoveType
from sharpy.plans import BuildOrder, SequentialList, Step, StepBuildGas
from sharpy.plans.acts import ActBuilding, Tech, Expand
from sharpy.plans.acts.zerg import AutoOverLord, MorphLair, ZergUnit
from sharpy.plans.require import All, Gas, UnitExists
from sharpy.plans.tactics import (
    PlanDistributeWorkers,
    PlanFinishEnemy,
    PlanZoneAttack,
    PlanWorkerOnlyDefense,
    PlanZoneDefense,
    PlanZoneGather,
)
from paul_plans.operations import CustomInject, NydusTarget

if TYPE_CHECKING:
    from sharpy.knowledges import Knowledge


class NydusRush(BuildOrder):
    def __init__(self):
        self.nydus_rally_set = False
        self.gas_work = PlanDistributeWorkers(min_gas=3)

        opening = SequentialList(
            ZergUnit(UnitTypeId.DRONE, to_count=13, only_once=True),
            ZergUnit(UnitTypeId.OVERLORD, to_count=2, only_once=True),
            ZergUnit(UnitTypeId.DRONE, to_count=15, only_once=True),
            Expand(2, consider_worker_production=False, priority=True),
            ZergUnit(UnitTypeId.DRONE, to_count=16, only_once=True),
            StepBuildGas(1),
            ActBuilding(UnitTypeId.SPAWNINGPOOL),
            ZergUnit(UnitTypeId.DRONE, to_count=19, only_once=True),
            ZergUnit(UnitTypeId.OVERLORD, to_count=3, only_once=True),
            ZergUnit(UnitTypeId.DRONE, to_count=22, only_once=True),
            Step(UnitExists(UnitTypeId.SPAWNINGPOOL, include_pending=False), ZergUnit(UnitTypeId.QUEEN, to_count=2)),
            Step(UnitExists(UnitTypeId.SPAWNINGPOOL, include_pending=False), ZergUnit(UnitTypeId.ZERGLING, to_count=2)),
            Step(UnitExists(UnitTypeId.LAIR, include_pending=False), ActBuilding(UnitTypeId.NYDUSNETWORK)),
        )

        unit_spam = BuildOrder(
            AutoOverLord(),
            ZergUnit(UnitTypeId.ZERGLING),
            Step(UnitExists(UnitTypeId.NYDUSNETWORK, include_pending=True), ZergUnit(UnitTypeId.QUEEN)),
        )

        upgrades = SequentialList(
            Step(
                All(Gas(100), UnitExists(UnitTypeId.SPAWNINGPOOL, include_pending=False)),
                Tech(UpgradeId.ZERGLINGMOVEMENTSPEED),
            ),
            Step(Gas(100), MorphLair()),
        )

        self.attack = PlanZoneAttack(20)
        self.nydus_attack = PlanZoneAttack(1)
        attack_tactics = SequentialList(
            PlanWorkerOnlyDefense(),
            PlanZoneDefense(),
            PlanZoneGather(),
            Step(
                None,
                self.attack,
                skip=UnitExists(UnitTypeId.NYDUSCANAL, include_pending=False, include_not_ready=False),
            ),
            Step(
                None,
                self.nydus_attack,
                skip_until=UnitExists(UnitTypeId.NYDUSCANAL, include_pending=False, include_not_ready=False),
            ),
            PlanFinishEnemy(),
        )

        super().__init__([opening, upgrades, unit_spam, self.gas_work, CustomInject(), NydusTarget(), attack_tactics])

    async def start(self, knowledge: "Knowledge"):
        await super(NydusRush, self).start(knowledge)

    async def execute(self) -> bool:
        if not self.nydus_rally_set:
            self.set_nydus_rally()
        self.manage_gas()
        self.network_load_or_unload_all()
        self.canal_unload_all()
        self.catch_nearby_units()
        self.catch_offensive_units_exiting()

        if self.get_count(UnitTypeId.NYDUSCANAL, include_pending=False, include_not_ready=False):
            self.combat.add_units(self.cache.own({UnitTypeId.ZERGLING, UnitTypeId.QUEEN}))
            target = self.attack._get_target()
            if target:
                self.combat.execute(target, MoveType.Assault)
            self.roles.set_tasks(UnitTask.Attacking, self.cache.own({UnitTypeId.ZERGLING, UnitTypeId.QUEEN}))

        return await super(NydusRush, self).execute()

    def set_nydus_rally(self):
        if self.cache.own(UnitTypeId.NYDUSNETWORK):
            for hatch in self.ai.townhalls:
                self.ai.do(
                    hatch(AbilityId.RALLY_HATCHERY_UNITS, self.cache.own(UnitTypeId.NYDUSNETWORK).first.position)
                )
            self.nydus_rally_set = True

    def manage_gas(self):
        if self.get_count(UnitTypeId.NYDUSNETWORK) and self.ai.vespene >= 200:
            self.gas_work.min_gas = 0
            self.gas_work.max_gas = 0
        else:
            self.gas_work.min_gas = 3
            self.gas_work.max_gas = 3

    def network_load_or_unload_all(self):
        load = True

        if self.get_count(UnitTypeId.NYDUSNETWORK, include_pending=False):
            for zone in self.zone_manager.expansion_zones:
                if zone.is_ours and zone.is_under_attack:
                    load = False
                    break
            if self.cache.own(UnitTypeId.NYDUSNETWORK).ready:
                network = self.cache.own(UnitTypeId.NYDUSNETWORK).ready.first
                if load:
                    idle_ling_queen = self.roles.get_types_from({UnitTypeId.ZERGLING, UnitTypeId.QUEEN}, UnitTask.Idle)
                    for u in idle_ling_queen.closer_than(30, network.position):
                        self.ai.do(network(AbilityId.LOAD_NYDUSNETWORK, u))
                else:
                    self.ai.do(network(AbilityId.UNLOADALL_NYDASNETWORK))

    def canal_unload_all(self):
        for canal in self.cache.own(UnitTypeId.NYDUSCANAL).ready:
            self.ai.do(canal(AbilityId.UNLOADALL_NYDUSWORM))

    def catch_offensive_units_exiting(self):
        idle_ling_queen = self.roles.get_types_from({UnitTypeId.ZERGLING, UnitTypeId.QUEEN}, UnitTask.Idle)
        for canal in self.cache.own(UnitTypeId.NYDUSCANAL).ready:
            close = self.cache.own([UnitTypeId.ZERGLING, UnitTypeId.QUEEN]).closer_than(2, canal.position)
            for u in close:
                if u in idle_ling_queen:
                    self.combat.add_unit(u)

    def catch_nearby_units(self):
        if self.cache.own(UnitTypeId.NYDUSNETWORK) and self.cache.own(UnitTypeId.NYDUSCANAL):
            network = self.cache.own(UnitTypeId.NYDUSNETWORK).first
            for unit in self.cache.own([UnitTypeId.ZERGLING, UnitTypeId.QUEEN]):
                if unit.distance_to(network.position) < 5:
                    self.ai.do(network(AbilityId.LOAD_NYDUSNETWORK, unit))
