from typing import TYPE_CHECKING

from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId
from sharpy.managers.build_detector import EnemyRushBuild
from sharpy.plans import BuildOrder, SequentialList, Step, StepBuildGas
from sharpy.plans.acts import ActBuilding, AutoWorker, Expand, Tech, DefensiveBuilding, DefensePosition
from sharpy.plans.acts.zerg import AutoOverLord, MorphHive, MorphLair, ZergUnit
from sharpy.plans.require import TechReady, Time, UnitExists
from sharpy.plans.tactics import (
    PlanDistributeWorkers,
    PlanFinishEnemy,
    PlanWorkerOnlyDefense,
    PlanZoneAttack,
    PlanZoneDefense,
    PlanZoneGather,
)
from sharpy.plans.tactics.scouting import Scout, ScoutLocation
from sharpy.plans.tactics.zerg import PlanHeatOverseer, InjectLarva, SpreadCreep
from sharpy.plans.tactics.zone_attack import AttackStatus
from paul_plans.compositions import OldLingBane
from paul_plans.operations import CustomInject
from paul_plans.opening_template import GenericMacroOpening

# from paul_plans.operations import SpreadCreepV2

if TYPE_CHECKING:
    from sharpy.knowledges import Knowledge


class MacroBuild(BuildOrder):
    def __init__(self):
        opening = GenericMacroOpening()

        game_plan = SequentialList(
            opening,
            Step(None, LingBaneUltra(), skip=lambda k: k.build_detector.rush_build == EnemyRushBuild.RoachRush),
            Step(
                None, RoachRushDefense(), skip_until=lambda k: k.build_detector.rush_build == EnemyRushBuild.RoachRush
            ),
        )

        upgrades = SequentialList(
            Step(UnitExists(UnitTypeId.SPAWNINGPOOL), Tech(UpgradeId.ZERGLINGMOVEMENTSPEED)),
            Step(None, Tech(UpgradeId.OVERLORDSPEED)),
        )

        scouting = SequentialList(
            Step(None, Scout(UnitTypeId.OVERLORD, 1, ScoutLocation.scout_main()), skip_until=Time(3 * 60)),
            Step(None, Scout(UnitTypeId.OVERLORD, 1, ScoutLocation.scout_main()), skip_until=Time(6 * 60)),
            Step(None, Scout(UnitTypeId.OVERLORD, 1, ScoutLocation.scout_main()), skip_until=Time(9 * 60)),
            Step(None, Scout(UnitTypeId.OVERLORD, 1, ScoutLocation.scout_main()), skip_until=Time(12 * 60)),
        )

        self.worker_distribution = PlanDistributeWorkers()

        auto_macro = BuildOrder(
            self.worker_distribution,
            AutoOverLord(),
            Step(None, AutoWorker(to_count=85), skip_until=self.should_drone),
            CustomInject(),
            SpreadCreep(),
        )

        overseers = ZergUnit(UnitTypeId.OVERSEER, to_count=3)

        super(MacroBuild, self).__init__(
            auto_macro, game_plan, upgrades, overseers, scouting,
        )

    def should_drone(self, _knowledge):
        """
        Aggressively drone to 60, then slightly less aggressively to 85.
        :param _knowledge:
        :return:
        """
        for zone in self.knowledge.expansion_zones:
            if zone.is_ours and zone.is_under_attack:
                return False
        # if self.ai.supply_workers < 60:
        #     if self.knowledge.game_analyzer.predicting_defeat:
        #         return False
        # else:
        #     if self.knowledge.game_analyzer.army_at_least_clear_disadvantage:
        #         return False
        if self.knowledge.game_analyzer.predicting_defeat:
            return False
        return True


class LingBaneUltra(BuildOrder):
    """
    Spam lings, banes up to half ling count, 15 ultras.
    """

    def __init__(self):
        self.queen = ZergUnit(UnitTypeId.QUEEN, 3)
        self.baneling = ZergUnit(UnitTypeId.BANELING, to_count=20)
        self.ultra = ZergUnit(UnitTypeId.ULTRALISK, to_count=15, priority=True)

        buildings = SequentialList(
            ActBuilding(UnitTypeId.SPAWNINGPOOL),
            ActBuilding(UnitTypeId.BANELINGNEST),
            MorphLair(),
            Step(UnitExists(UnitTypeId.HATCHERY, count=3), ActBuilding(UnitTypeId.EVOLUTIONCHAMBER, to_count=2)),
            Step(TechReady(UpgradeId.ZERGMELEEWEAPONSLEVEL2, 0.2), ActBuilding(UnitTypeId.INFESTATIONPIT)),
            Step(UnitExists(UnitTypeId.INFESTATIONPIT, include_pending=False), MorphHive()),
            Step(UnitExists(UnitTypeId.HIVE), ActBuilding(UnitTypeId.ULTRALISKCAVERN)),
        )

        self.expander = Expand(3, priority=True, consider_worker_production=False)
        self.gas = StepBuildGas(3)

        self.attack = PlanZoneAttack(120)
        attack_tactics = [
            PlanHeatOverseer(),
            PlanWorkerOnlyDefense(),
            PlanZoneGather(),
            PlanZoneDefense(),
            self.attack,
            PlanFinishEnemy(),
        ]

        pool_upgrades = SequentialList(
            Tech(UpgradeId.ZERGLINGMOVEMENTSPEED),
            Step(UnitExists(UnitTypeId.HIVE, include_pending=False), Tech(UpgradeId.ZERGLINGATTACKSPEED),),
        )

        ultra_upgrades = SequentialList(
            Tech(UpgradeId.CHITINOUSPLATING),
            Tech(UpgradeId.ANABOLICSYNTHESIS),
        )

        nest_upgrades = Step(UnitExists(UnitTypeId.LAIR), Tech(UpgradeId.CENTRIFICALHOOKS))

        evo_upgrades = SequentialList(
            Tech(UpgradeId.ZERGMELEEWEAPONSLEVEL1),
            Step(None, Tech(UpgradeId.ZERGGROUNDARMORSLEVEL1), skip_until=UnitExists(UnitTypeId.EVOLUTIONCHAMBER, 2)),
            Step(UnitExists(UnitTypeId.LAIR, include_pending=False), Tech(UpgradeId.ZERGMELEEWEAPONSLEVEL2)),
            Step(
                UnitExists(UnitTypeId.LAIR, include_pending=False),
                Tech(UpgradeId.ZERGGROUNDARMORSLEVEL2),
                skip_until=UnitExists(UnitTypeId.EVOLUTIONCHAMBER, 2),
            ),
            Step(UnitExists(UnitTypeId.HIVE, include_pending=False), Tech(UpgradeId.ZERGMELEEWEAPONSLEVEL3)),
            Step(
                UnitExists(UnitTypeId.HIVE, include_pending=False),
                Tech(UpgradeId.ZERGGROUNDARMORSLEVEL3),
                skip_until=UnitExists(UnitTypeId.EVOLUTIONCHAMBER, 2),
            ),
        )

        units = BuildOrder(
            self.queen,
            Step(UnitExists(UnitTypeId.ULTRALISKCAVERN, include_pending=False), self.ultra),
            self.baneling,
            ZergUnit(UnitTypeId.ZERGLING),
        )

        super(LingBaneUltra, self).__init__(
            self.expander,
            buildings,
            pool_upgrades,
            nest_upgrades,
            evo_upgrades,
            ultra_upgrades,
            units,
            self.gas,
            attack_tactics,
        )

    async def execute(self) -> bool:
        hatches = self.get_count(UnitTypeId.HATCHERY, include_pending=False)
        if 3 <= hatches < 5:
            if self.ai.minerals >= 300 and not self.cache.own(UnitTypeId.LARVA).amount:
                self.expander.to_count = hatches + 1
        else:
            zone_minerals = 0
            for zone in self.knowledge.expansion_zones:
                if zone.is_ours:
                    zone_minerals += zone.last_minerals
            if zone_minerals < 20000:
                self.expander.to_count = hatches + 1

        self.queen.to_count = self.get_count(UnitTypeId.HATCHERY) + 3
        self.baneling.to_count = self.cache.own(UnitTypeId.ZERGLING).amount // 2
        self.gas.to_count = (hatches - 1) * 2
        # can't alter priority, so we're changing the target number. If attacking, ling bane, otherwise, ultra first
        if self.attack.status == AttackStatus.Attacking:
            self.ultra.to_count = 0
        else:
            self.ultra.to_count = 15

        return await super(LingBaneUltra, self).execute()


class RoachRushDefense(BuildOrder):
    def __init__(self):
        buildings = SequentialList(
            ActBuilding(UnitTypeId.SPAWNINGPOOL),
            Step(
                UnitExists(UnitTypeId.SPAWNINGPOOL),
                DefensiveBuilding(UnitTypeId.SPINECRAWLER, DefensePosition.Entrance, to_count=3, to_base_index=1),
            ),
            ActBuilding(UnitTypeId.ROACHWARREN),
        )

        units = BuildOrder(
            Step(UnitExists(UnitTypeId.ROACHWARREN), ZergUnit(UnitTypeId.ROACH)),
            Step(UnitExists(UnitTypeId.SPAWNINGPOOL), ZergUnit(UnitTypeId.QUEEN, to_count=2)),
            Step(UnitExists(UnitTypeId.SPAWNINGPOOL), ZergUnit(UnitTypeId.ZERGLING)),
        )

        super(RoachRushDefense, self).__init__(buildings, units)
