from typing import List, TYPE_CHECKING
import numpy as np
from scipy.spatial import distance_matrix
from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.position import Point2
from sc2.units import Units

from sharpy.general.zone import Zone
from sharpy.plans.acts import ActBase
from sharpy.managers.roles import UnitTask

if TYPE_CHECKING:
    from sharpy.knowledges import Knowledge


class BanelingDropBase(ActBase):
    """
    Act for dropping banelings on targets. Requires Lair and Baneling nest.
    """

    dropperlords: Units
    banelings: Units
    drop_position: Point2
    retreat_position: Point2

    def __init__(self, drop_size: int = 1, enemy_expansion: int = 0):
        super().__init__()
        self.drop_size = drop_size
        self.started = False
        self.ended = False
        self.dropperlord_tags: List[int] = []
        self.baneling_tags: List[int] = []
        self.done_tags: List[int] = []
        self.enemy_expansion_target = enemy_expansion

    async def start(self, knowledge: "Knowledge"):
        await super().start(knowledge)
        self.dropperlords = Units([], self.ai)
        self.banelings = Units([], self.ai)
        target_base: Zone = self.knowledge.enemy_expansion_zones[self.enemy_expansion_target]
        self.drop_position = target_base.behind_mineral_position_center.towards(
            target_base.center_location, 3
        )
        self.retreat_position = self.find_retreat_position()

    async def execute(self) -> bool:
        # Drop is finished, skip
        if self.ended:
            return True

        # pause while we create Transport Overlords
        if self.create_dropperlords():
            return True

        # pause while we create banelings
        if self.create_banelings():
            return True

        self.dropperlords.clear()
        self.banelings.clear()

        self.find_units()

        if self.dropperlords:
            self.roles.set_tasks(UnitTask.Scouting, self.dropperlords)
            if self.banelings:
                self.roles.set_tasks(UnitTask.Reserved, self.banelings)
            await self.micro_units()
        return True

    def find_units(self):
        if not self.started:
            free_dropperlords = self.roles.get_types_from(
                UnitTypeId.OVERLORDTRANSPORT, UnitTask.Idle, UnitTask.Moving, UnitTask.Gathering, UnitTask.Reserved
            )
            free_banelings = self.roles.get_types_from(
                UnitTypeId.BANELING, UnitTask.Idle, UnitTask.Moving, UnitTask.Gathering
            )
            if free_dropperlords.amount >= self.drop_size and free_banelings.amount >= 4 * self.drop_size:
                new_dropperlords = free_dropperlords.random_group_of(self.drop_size)
                self.dropperlords.extend(new_dropperlords)
                self.dropperlord_tags = new_dropperlords.tags

                new_banelings = free_banelings.random_group_of(4 * self.drop_size)
                self.banelings.extend(new_banelings)
                self.baneling_tags = new_banelings.tags

                self.started = True
        else:
            dropperlords = self.roles.get_types_from(UnitTypeId.OVERLORDTRANSPORT, UnitTask.Scouting)
            self.dropperlords.extend(dropperlords.tags_in(self.dropperlord_tags))

            banelings = self.roles.get_types_from(UnitTypeId.BANELING, UnitTask.Reserved)
            self.banelings.extend(banelings.tags_in(self.baneling_tags))

            if not dropperlords and not banelings:
                self.ended = True
                return True

    async def micro_units(self) -> bool:
        for lord in self.dropperlords:
            if len(lord.passengers_tags) < 4:
                for bane in self.banelings.random_group_of(4 - len(lord.passengers_tags)):
                    self.ai.do(bane.move(lord.position))
                    self.ai.do(lord(AbilityId.LOAD_OVERLORD, bane))
            else:
                # we've arrived or are dying
                if lord.distance_to(self.drop_position) <= .5 or lord.health_percentage < .25:
                    self.ai.do(lord(AbilityId.UNLOADALLAT_OVERLORD, lord.position))
                    self.ai.do(lord.move(self.retreat_position, queue=True))
                    self.ai.do(lord.move(self.knowledge.expansion_zones[1].gather_point, queue=True))
                    self.done_tags.append(lord.tag)
                else:
                    self.ai.do(lord.move(self.drop_position))
        if len(self.done_tags) >= self.drop_size:
            self.ended = True

    def create_dropperlords(self) -> bool:
        if self.cache.own([UnitTypeId.OVERLORDTRANSPORT, UnitTypeId.TRANSPORTOVERLORDCOCOON]).amount < self.drop_size:
            free_overlords = self.roles.get_types_from(
                {UnitTypeId.OVERLORD}, UnitTask.Idle, UnitTask.Moving, UnitTask.Gathering, UnitTask.Reserved
            )
            if free_overlords:
                for lord in free_overlords.random_group_of(
                    self.drop_size
                    - self.cache.own([UnitTypeId.OVERLORDTRANSPORT, UnitTypeId.TRANSPORTOVERLORDCOCOON]).amount
                ):
                    self.ai.do(lord(AbilityId.MORPH_OVERLORDTRANSPORT))
                return True
        return False

    def create_banelings(self) -> bool:
        if self.cache.own(UnitTypeId.BANELING).amount < 4 * self.drop_size:
            free_lings = self.roles.get_types_from(
                {UnitTypeId.ZERGLING}, UnitTask.Idle, UnitTask.Moving, UnitTask.Gathering
            )
            for ling in free_lings.random_group_of(
                    (4 * self.drop_size) - self.cache.own(UnitTypeId.BANELING).amount
            ):
                self.ai.do(ling(AbilityId.MORPHZERGLINGTOBANELING_BANELING))
            return True
        return False

    def find_retreat_position(self) -> Point2:
        overlord_positions = [tuple(x) for x in self.knowledge.pathing_manager.overlord_spots]
        distances = distance_matrix([self.drop_position], overlord_positions)[0]
        closest_spot = overlord_positions[np.where(distances == min(distances))[0][0]]
        return Point2(closest_spot)
