"""Put a Nydus in their main."""
from typing import List, TYPE_CHECKING
from sc2.position import Point2
from sharpy.plans.acts import ActBase
from sharpy.managers.roles import UnitTask
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.ability_id import AbilityId
from sc2.unit import Unit
from sc2.units import Units
import numpy as np
from scipy import spatial

if TYPE_CHECKING:
    from sharpy.knowledges import Knowledge


class NydusTarget(ActBase):
    """
    Targets hidden spot in main base unless given an expansion number.
    """
    nydus_location: Point2
    retreat_location: Point2

    def __init__(self, enemy_expansion_target: int = 0):
        self.overlord: Unit = None
        self.placing_nydus = False
        self.expansion_target = enemy_expansion_target
        super(NydusTarget, self).__init__()

    async def start(self, knowledge: "Knowledge"):
        await super(NydusTarget, self).start(knowledge)
        if self.expansion_target == 0:
            self.nydus_location = self.calc_nydus_spot()
        else:
            self.nydus_location = self.knowledge.enemy_expansion_zones[self.expansion_target].gather_point
        self.retreat_location = self.calc_retreat_location()

    async def execute(self) -> bool:
        if self.get_count(UnitTypeId.NYDUSCANAL):
            self.ai.do(self.overlord.move(self.retreat_location))
        self.overlord = self.roles.get_types_from({UnitTypeId.OVERLORD}, UnitTask.Hallucination)
        if not self.overlord:
            ovies = self.cache.own(UnitTypeId.OVERLORD)
            if not ovies:
                return True
            self.overlord = ovies.closest_to(self.nydus_location)
            self.roles.set_task(UnitTask.Hallucination, self.overlord)
        if isinstance(self.overlord, Units):
            self.overlord = self.overlord.closest_to(self.nydus_location)
        if isinstance(self.overlord, Unit):
            self.move_overlord(self.overlord)
        if (
            self.get_count(UnitTypeId.NYDUSNETWORK, include_pending=False, include_not_ready=False)
            and not self.get_count(UnitTypeId.NYDUSCANAL, include_not_ready=True, include_pending=True)
            and self.ai.can_afford(UnitTypeId.NYDUSCANAL)
        ):
            self.placing_nydus = True
            network = self.cache.own(UnitTypeId.NYDUSNETWORK).random
            for i in range(11):
                pos = self.nydus_location.towards(self.knowledge.enemy_main_zone.center_location, i)
                if await self.ai.can_place(UnitTypeId.NYDUSCANAL, pos) and self.ai.is_visible(pos):
                    self.ai.do(network(AbilityId.BUILD_NYDUSWORM, pos))
                    break
                else:
                    if i == 10:
                        self.ai.do(self.overlord.move(self.overlord.position.towards(self.nydus_location, 1)))
        return True

    def calc_nydus_spot(self):
        """
        Find the location in the enemy main furthest from the townhall and at least 9 distance away from gas geysers.
        """
        # create KDTree for nearest neighbor searching composed only of points within the pathing grid
        standin = [
            (x, y)
            for x in range(self.knowledge.ai.game_info.terrain_height.width)
            for y in range(self.knowledge.ai.game_info.terrain_height.height)
        ]
        tree = spatial.KDTree(standin)

        # find enemy main gas geysers and unique points within 9 of them (the typical sight radius of a gas building)
        enemy_gases = self.knowledge.ai.vespene_geyser.closer_than(12, self.knowledge.enemy_main_zone.center_location)
        gas_one = [standin[x] for x in tree.query_ball_point(enemy_gases[0].position, 9)]
        gas_two = [standin[y] for y in tree.query_ball_point(enemy_gases[1].position, 9)]
        close_to_gas = set(gas_one + gas_two)

        # find points that are too close to the ramp
        close_to_ramp = [standin[z] for z in tree.query_ball_point(self.knowledge.enemy_main_zone.ramp.top_center, 13)]

        # get the height of the enemy main
        enemy_main_height = self.knowledge.ai.game_info.terrain_height.data_numpy[
            self.knowledge.enemy_main_zone.center_location.rounded[1],
            self.knowledge.enemy_main_zone.center_location.rounded[0],
        ]

        # "find" the enemy main tiles based on tiles within distance that are the same height as the main
        terrain = self.knowledge.ai.game_info.terrain_height
        pathing = self.knowledge.ai.game_info.pathing_grid
        enemy_main = [
            standin[z]
            for z in tree.query_ball_point(self.knowledge.enemy_main_zone.center_location, 45)
            if terrain[standin[z]] == enemy_main_height and pathing[standin[z]] == 1
        ]

        # remove points that are too close to the gas geysers and ramp
        main_away_from_gas_and_ramp = list(set(enemy_main) - close_to_gas - set(close_to_ramp))

        # compute the distances of each remaining point to the enemy townhall
        distances = spatial.distance_matrix(
            main_away_from_gas_and_ramp, [self.knowledge.enemy_main_zone.center_location]
        )

        # select the first point furthest from the enemy townhall
        nydus_raw_location = main_away_from_gas_and_ramp[np.where(distances == max(distances))[0][0]]

        # return the location as Point2
        return Point2(nydus_raw_location)

    def get_enemy_main(self, standin, tree) -> List:
        terrain = self.knowledge.ai.game_info.terrain_height
        pathing = self.knowledge.ai.game_info.pathing_grid
        enemy_main_base = [
            standin[z]
            for z in tree.query_ball_point(self.knowledge.enemy_main_zone.center_location, 45)
            if terrain[standin[z]] == 223 and pathing[standin[z]] == 1  # highest elevation in maps is 223
        ]
        return enemy_main_base

    def calc_retreat_location(self):
        if not self.nydus_location:
            # if we couldn't find a nydus location, send home afterwards
            return self.knowledge.expansion_zones[0].center_location
        ol_spots = [(p.x, p.y) for p in self.knowledge.pathing_manager.overlord_spots]
        if ol_spots:
            # distances from the nydus location to the overlord spots
            retreat_distances = spatial.distance_matrix(ol_spots, [self.nydus_location])
            # get closest overlord spot
            target_location = ol_spots[np.where(retreat_distances == min(retreat_distances))[0][0]]
            # return location as Point2
            return Point2(target_location)
        else:
            # map didn't return any overlord spots, go home
            return self.knowledge.expansion_zones[0].center_location

    def move_overlord(self, unit: Unit):
        # if self.get_count(UnitTypeId.NYDUSCANAL, include_not_ready=True, include_pending=True):
        #     if unit.orders:
        #         self.ai.do(unit.stop())
        # if not self.placing_nydus and unit.distance_to(self.nydus_location) <= 10:
        #     if unit.orders:
        #         self.ai.do(unit.stop())
        # else:
        if not self.get_count(UnitTypeId.NYDUSCANAL, include_pending=True, include_not_ready=True, include_killed=True):
            if not unit.orders:
                if self.expansion_target == 0:
                    self.ai.do(unit.move(self.nydus_location.towards(self.knowledge.enemy_main_zone.center_location, -9)))
                else:
                    self.ai.do(unit.move(self.retreat_location))
