from typing import List, Tuple
from sc2.position import Point2
from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score


class OverlordSpots:
    """
    Calculate Overlord hiding places.
    """
    def __init__(self, info):
        self.game_info = info
        self.spots = self.find_overlord_spots()

    def find_overlord_spots(self) -> List[Point2]:
        """
        Find clusters of high ground tiles for overlords
        :return List[List[float]]: List of Overlord locations as Point2
        """
        # create pathing grid tiles -> filter so that only unpathable highground tiles with no pathable neighbors remain
        all_tiles = [
            (x, y)
            for x in range(self.game_info.pathing_grid.width)
            for y in range(self.game_info.pathing_grid.height)
        ]
        # PillarsOfGold has one tile at 228 and the mains at 223
        # max_height = self.game_info.terrain_height.data_numpy.max()

        main_height = self.game_info.terrain_height[self.game_info.player_start_location.rounded]

        high_ground_tiles = [
            tile for tile in all_tiles
            if self.game_info.pathing_grid[tile] == 0
            and self.game_info.terrain_height[tile] >= main_height
        ]
        options = [
            tile for tile in high_ground_tiles
            if not any([self.game_info.pathing_grid[i] for i in Point2(tile).neighbors8])
        ]
        # determine the optimal number of clusters for KMeans using Davies-Bouldin index
        optimal_cluster_number = self.find_optimal_index(options)
        # find clusters of tiles
        kmeans = KMeans(n_clusters=optimal_cluster_number, random_state=0).fit(options)
        # return center points of clusters as Point2
        overlord_spots = [Point2(loc) for loc in kmeans.cluster_centers_]
        return overlord_spots

    # noinspection PyMethodMayBeStatic
    def find_optimal_index(self, tiles: List[Tuple]) -> int:
        """
        Use Davies-Bouldin score to determine optimal cluster number
        :param tiles: List of Tuples representing high ground locations
        :return: int
        """
        db_dict = {}
        if len(tiles) < 8:
            return len(tiles)
        for i in range(8, 40):
            kmeans = KMeans(n_clusters=i, random_state=0).fit(tiles)
            if i == len(kmeans.labels_):
                # davies_bouldin_score accepts 2 to n_samples (len(labels_)) - 1 (inclusive)
                break
            db_dict[davies_bouldin_score(tiles, kmeans.labels_)] = i

        return db_dict[min(db_dict)]
